//

// = FILENAME
//    PRM.cc
//
// = AUTHOR(S)
//    Patric Jensfelt
//
/*----------------------------------------------------------------------*/

#include "PRM.h"
#include <cstdlib>
#include <iostream>
#include <cfloat>

PRM::PRM(unsigned int n_o)
{
  // Make sure that the random generator has been initialize
  srand(time(0));
  no_arms = n_o;
}

PRM::~PRM()
{
}

int 
PRM::buildRoadMap(World &world, 
                  int nNodes, int K, double step)
{
  // Store step size so that the same value can be used in findPath
  m_Step = step;

  clearNodes();

  // Clear the vector with edges (used for debug display stuff)
  m_Edges.clear();

  // Create new set of n nodes with random positions. Each new node is
  // first checked for collisions before being added
  int n = 0;
  while (m_Nodes.size() < nNodes) {

    PRMNode *node = generateNodeWithRandomPosition();

    // Check if we can add it, is it does not collide with anything
    if (world.collidesWith(node->geta())) {

      // Delete the newly created node to avoid memory leaks
      delete node;

      // Go back up and create a new one
      continue;
    }

    // Try to connect this node to existing nodes in the neighborhod
    // of this node
    
    // Get list of distances to other nodes
    std::list< std::pair<double, int> > distances;
    getNodeDistances(node, distances);
    
    int k = 0;
    for (std::list< std::pair<double, int> >::iterator iter = distances.begin();iter != distances.end(); iter++) {
      // Skip the nodes if it is the same as we try to connect to
      if (iter->second == n) {
        continue;
      }

      // Check if the path between the nodes is free from collisions
      if (isCollisionFreePath(world, node->geta(), m_Nodes[iter->second]->geta(), step)) {
        m_Nodes[iter->second]->addNewEdge(node);
        node->addNewEdge(m_Nodes[iter->second]);
        m_Edges.push_back( std::make_pair(n, iter->second) );        
      }
      
      // If we have tested enough of the neighbors we break here
      k++;
      if (k >= K) break;
    } 

    m_Nodes.push_back(node);
    n++;
  }
}

void
PRM::clearNodes()
{
  // Clear nodes
  for (unsigned int i = 0; i < m_Nodes.size(); i++) 
    delete m_Nodes[i];
  m_Nodes.clear();
}

void
PRM::getNodeDistances(PRMNode *n, std::list< std::pair<double, int> > &distances)
{
  unsigned int i, j;
  double dt, dis, *tempi, *tempn=n->geta();
  for (i = 0; i < m_Nodes.size(); i++) {
    dis=0;
    tempi = m_Nodes[i]->geta();
    for(j=0;j<no_arms;j++) {
      dt = tempi[j] - tempn[j];
      if(dt < -180.0)
	dt=dt+360.0;
      else if (dt > 180.0)
	dt=dt-360.0;
      dis += dt*dt;
    }
    dis = sqrt(dis);
    distances.push_back( std::make_pair(dis, i) );
  }

  // Sort the list with distances so that the closest one comes first
  distances.sort();
}

bool
PRM::isCollisionFreePath(World &world, double sa[], double ea[], double step)
{
  int k;
  double dt[MAX_ARMS], dis=0, ste[MAX_ARMS];
  for(k=0;k<no_arms;k++) {
    dt[k] = ea[k]-sa[k];
    if(dt[k] < -180.0)
      dt[k] += 360.0;
    else if (dt[k] > 180.0)
      dt[k] -= 360.0;
    dis += dt[k]*dt[k];
  }
  if(dis == 0)
    return false;
  dis = sqrt(dis);
  for(k=0;k<no_arms;k++) {
    ste[k] = dt[k]/dis;
  }
  bool last = false;
  bool done = false;
  double pos = 0; // where on the line are we
  double xv[MAX_ARMS];
  while (!done) {
    
    for(k=0;k<no_arms;k++) 
      xv[k] = sa[k] + pos * ste[k];
    
    if (world.collidesWith(xv)) return false;
 
    if (!last && pos > dis) {
      last = true;
      pos = dis;
    } else if (last) {
      done = true; 
    }
    pos += step;
  }

  return true;
}

int
PRM::findClosestNodeOnGraph(World &world, double as[])
{
  double minD = DBL_MAX;
  int minI = -1;
  int k;
  double dt[MAX_ARMS], dis=0;
  for (unsigned int i = 0; i < m_Nodes.size(); i++) {
    dis=0;
    for(k=0;k<no_arms;k++) {
      dt[k] = as[k]-m_Nodes[i]->geta()[k];
      if(dt[k] < -180.0)
	dt[k] += 360.0;
      else if (dt[k] > 180.0)
	dt[k] -= 360.0;
      dis += dt[k]*dt[k];
    }
    dis = sqrt(dis);
    if (dis < minD &&
        isCollisionFreePath(world,
			    as, m_Nodes[i]->geta(),
                            m_Step)) {
      minD = dis;
      minI = i;
    }
  }
 
  return minI;
}

bool 
PRM::findPath(World &world,
	      double starta[], double enda[],
              std::list<Position> &path)
{
  path.clear();

  // Find the node in the graf that is closest to the start and goal
  // point and offer a collision free path from these points to the
  // road map.
  if(world.collidesWith(starta) ||  world.collidesWith(enda)) {
    std::cerr<<"Start or Goal States not valid\n";
    return false;
  }
  int start = findClosestNodeOnGraph(world, starta);
  int goal = findClosestNodeOnGraph(world, enda);
  if (start < 0) {
    std::cerr << "Could not connect the start point "
              << "to the road map. You should think about rebuilding it\n";
    return false;
  }
  if (goal < 0) {
    std::cerr << "Could not connect the goal point ("
              << "to the road map. You should think about rebuilding it\n";
    return false;
  }

  double pathLen = 0;
  if (m_Nodes[goal]->findPath(m_Nodes[start], pathLen)) {

    path.push_back(Position(no_arms, starta));

    PRMNode *n = m_Nodes[start];

    while (n) {

      path.push_back( Position( ((PRMNode*)n)->getn(), 
                                ((PRMNode*)n)->geta() ) );

      n = (PRMNode*)n->parent;
    }

    path.push_back(Position(no_arms, enda));

    return true;
  }

  return false;  
}

PRMNode* 
PRM::generateNodeWithRandomPosition()
{
  PRMNode* n = new PRMNode();
  n->setn(no_arms);
  int k;
  for(k=0;k<no_arms;k++)
    n->angles[k]=rand()%360;
  return n;
}

// void 
// PRM::writeMatlabDisplayCode(std::ostream &s, bool dispEdges, bool dispNodes)
// {
//   if (dispNodes) {
//     for (unsigned int i = 0; i < m_Nodes.size(); i++) {
//       s << "plot(" 
//         << m_Nodes[i]->getX() << "," 
//         << m_Nodes[i]->getY() << ",\'.g\')\n";
//     }
//   }
  
//   // We always write the code for displaying the edges but in case we
//   // are ordered not to displayit we add a "if 0 ... end" around it
//   // that can be removed manually
//   s << "if " << (int)dispEdges << std::endl;
//   for (std::vector< std::pair<int, int> >::iterator i = m_Edges.begin();
//        i != m_Edges.end(); i++) {
//     s << "  plot([" 
//       << m_Nodes[i->first]->getX() << "," 
//       << m_Nodes[i->second]->getX() << "] ,[" 
//       << m_Nodes[i->first]->getY() << ", "
//       << m_Nodes[i->second]->getY() << "], \'k\')\n";
//   }
//   s << "end" << std::endl;
// }
